"use strict";
/* eslint-disable valid-jsdoc */
/**
 * penguins-eggs-v7
 * author: Piero Proietti
 * email: piero.proietti@gmail.com
 * license: MIT
 */
Object.defineProperty(exports, "__esModule", { value: true });
const shx = require("shelljs");
const fs = require("fs");
const dns = require("dns");
const path = require("path");
const os = require("os");
const pjson = require("pjson");
const inquirer = require("inquirer");
const chalk = require("chalk");
const pacman_1 = require("./pacman");
const clear = require("clear");
const figlet = require("figlet");
/**
 * Utils: general porpourse utils
 * @remarks all the utilities
 */
class Utils {
    /**
     * Restituisce il prefisso della iso
     * @param distroId
     * @param versionId
     */
    static snapshotPrefix(distroId, versionId) {
        let result = 'egg-of-' + distroId.toLowerCase() + '-' + versionId.toLowerCase() + '-';
        result = result.replace(`/`, '-');
        return result;
    }
    /**
    * Controlla se il sistema è avviato con systemd
    */
    static isSystemd() {
        // return (shx.exec(`pidof systemd`).stdout.trim() === '1')
        return (shx.exec(`ps -p 1 -o comm=`).stdout.trim() === 'systemd');
    }
    /**
     * ricava path per vmlinuz
     */
    static vmlinuz() {
        let results = fs.readFileSync('/proc/cmdline', 'utf8').split(' ');
        let result = results[0];
        result = result.substring(result.indexOf('=') + 1);
        if (result.indexOf('@') > 0) {
            result = result.substring(result.indexOf('@') + 1);
        }
        return result;
    }
    /**
     * ricava path per initrdImg
     */
    static initrdImg() {
        const vmlinuz = Utils.vmlinuz();
        const path = vmlinuz.substring(0, vmlinuz.lastIndexOf('/')) + '/';
        const version = vmlinuz.substring(vmlinuz.indexOf('-'));
        return path + 'initrd.img' + version;
    }
    /**
     * Occore vedere un modo per creare machine-id dove non esiste
     */
    static machineId() {
        let result = '';
        if (fs.existsSync('/etc/machine-id')) {
            result = fs.readFileSync('/etc/machine-id', 'utf-8').trim();
        }
        else if (fs.existsSync('/var/lib/dbus/machine-id')) {
            result = fs.readFileSync('/var/lib/dbus/machine-id', 'utf-8').trim();
        }
        return result;
    }
    /**
     *
     * @param msg
     */
    static warning(msg = '') {
        console.log('eggs >>> ' + chalk.cyanBright(msg) + '.');
    }
    static error(msg = '') {
        console.log('eggs >>> ' + chalk.bgGrey(msg) + '.');
    }
    /**
     * Return the primary user's name
     */
    static getPrimaryUser() {
        // let primaryUser = shx.exec(`echo $(awk -F":" '/1000:1000/ { print $1 }' /etc/passwd)`, { silent: true }).stdout.trim()
        const primaryUser = shx.exec('echo $SUDO_USER', { silent: true }).stdout.trim();
        if (primaryUser === '') {
            console.log('Cannot find your user name. Log as normal user and run: $sudo eggs [COMMAND]');
            process.exit(1);
        }
        return primaryUser;
    }
    /**
     * restituisce uuid
     * @param device
     */
    static uuid(device) {
        const uuid = shx.exec(`blkid -s UUID -o value ${device}`).stdout.trim();
        return uuid;
    }
    /**
     *
     * @param date
     */
    static formatDate(date) {
        const d = new Date(date);
        let month = String(d.getMonth() + 1);
        let day = String(d.getDate());
        const year = d.getFullYear();
        let hh = String(d.getHours());
        let mm = String(d.getMinutes());
        if (month.length < 2) {
            month = '0' + month;
        }
        if (day.length < 2) {
            day = '0' + day;
        }
        if (hh.length < 2) {
            hh = '0' + hh;
        }
        if (mm.length < 2) {
            mm = '0' + mm;
        }
        return [year, month, day].join('-') + '_' + hh + mm;
    }
    /**
     * return the name of the package: penguins-eggs
     * @returns penguins-eggs
     */
    static getPackageName() {
        return pjson.name;
    }
    /**
     * Count the eggs present in the nest
     * @returns {number} Numero degli snapshot presenti
     */
    static getSnapshotCount(snapshot_dir = '/') {
        if (fs.existsSync(snapshot_dir)) {
            const list = fs.readdirSync(snapshot_dir);
            if (list.length > 0) {
                return list.length - 1;
            }
        }
        return 0;
    }
    /**
     * Get the syze of the snapshot
     * @returns {string} grandezza dello snapshot in Byte
     */
    static getSnapshotSize(snapshot_dir = '/') {
        let fileSizeInBytes = 0;
        const size = shx.exec(`/usr/bin/find /home/eggs -maxdepth 1 -type f -name '*.iso' -exec du -sc {} + | tail -1 | awk '{print $1}'`, { silent: true }).stdout.trim();
        if (size === '') {
            fileSizeInBytes = 0;
        }
        else {
            fileSizeInBytes = Number(size);
        }
        return fileSizeInBytes;
    }
    /**
     * Return the eggName with architecture and date
     * @param basename
     * @returns eggName
     */
    static getFilename(basename = '') {
        let arch = 'x64';
        if (Utils.isi686()) {
            arch = 'i386';
        }
        let isoName = `${basename}-${arch}_${Utils.formatDate(new Date())}`;
        if (isoName.length >= 28) {
            isoName = isoName.substr(0, 28); // 28 +  4 .iso = 32 lunghezza max di volid
        }
        return `${isoName}.iso`;
    }
    /**
     * Calculate the space used on the disk
     * @return {void}
     */
    static getUsedSpace() {
        let fileSizeInBytes = 0;
        if (this.isLive()) {
            fileSizeInBytes = 0; // this.getLiveRootSpace()
        }
        else {
            fileSizeInBytes = Number(shx.exec(`df /home | /usr/bin/awk 'NR==2 {print $3}'`, {
                silent: true
            }).stdout);
        }
        return fileSizeInBytes;
    }
    /**
     * Extimate the linuxfs dimension
     * probably is better to rename it as
     * getLiveSpaceRootNeed
     * @returns {number} Byte
     */
    static getLiveRootSpace(type = 'debian-live') {
        let squashFs = '/run/live/medium/live/filesystem.squashfs';
        if (type === 'mx') {
            squashFs = '/live/boot-dev/antiX/linuxfs';
        }
        // Ottengo la dimensione del file compresso
        const compressedFs = fs.statSync(squashFs).size;
        // get compression factor by reading the linuxfs squasfs file, if available
        const compressedFs_compression_type = shx.exec(`dd if=${compressedFs} bs=1 skip=20 count=2 status=none 2>/dev/null| /usr/bin/od -An -tdI`);
        let compression_factor = 0;
        if (compressedFs_compression_type === '1') {
            compression_factor = 37; // gzip
        }
        else if (compressedFs_compression_type === '2') {
            compression_factor = 52; // lzo, not used by antiX
        }
        else if (compressedFs_compression_type === '3') {
            compression_factor = 52; // lzma, not used by antiX
        }
        else if (compressedFs_compression_type === '4') {
            compression_factor = 31; // xz
        }
        else if (compressedFs_compression_type === '5') {
            compression_factor = 52; // lz4
        }
        else {
            compression_factor = 30; // anything else or linuxfs not reachable (toram), should be pretty conservative
        }
        let rootfs_file_size = 0;
        const linuxfs_file_size = (Number(shx.exec('df /live/linux --output=used --total | /usr/bin/tail -n1').stdout.trim()) * 1024 * 100) / compression_factor;
        if (fs.existsSync('/live/persist-root')) {
            rootfs_file_size = Number(shx.exec('df /live/persist-root --output=used --total | /usr/bin/tail -n1').stdout.trim()) * 1024;
        }
        let rootSpaceNeeded;
        if (type === 'mx') {
            /**
             * add rootfs file size to the calculated linuxfs file size. Probaby conservative, as rootfs will likely have some overlap with linuxfs
             */
            rootSpaceNeeded = linuxfs_file_size + rootfs_file_size;
        }
        else {
            rootSpaceNeeded = linuxfs_file_size;
        }
        return rootSpaceNeeded / 1073741824.0; // Converte in GB
    }
    /**
     * Return true if i686 architecture
     * @remarks to move in Utils
     * @returns {boolean} true se l'architettura è i686
     * Cambiare con process.arch === 'ia32'
     */
    static isi686() {
        return process.arch === 'ia32';
    }
    /**
     * Check se la macchina ha grub adatto ad efi
     */
    static isUefi() {
        let isUefi = false;
        if (process.arch === 'x64') {
            // grub-efi-amd64-bin grub-efi-ia32-bin
            if (pacman_1.default.packageIsInstalled('grub-efi-amd64-bin')) {
                isUefi = true;
            }
        }
        return isUefi;
    }
    /**
     * Controlla se è un pacchetto deb
     */
    static isDebPackage() {
        let ret = false;
        if (process.execPath !== '/usr/bin/node') {
            ret = true;
        }
        return ret;
    }
    /**
     * Controlla se è un pacchetto sorgente
     */
    static isSources() {
        let ret = false;
        if (__dirname.substring(0, 6) === '/home/') {
            ret = true;
        }
        return ret;
    }
    /**
     * Controlla se è un pacchetto npm
     */
    static isNpmPackage() {
        return !(this.isDebPackage() || this.isSources());
    }
    /**
     *
     */
    static rootPenguin() {
        return path.resolve(__dirname, '../../');
    }
    /**
     * return the short name of the package: eggs
     * @returns eggs
     */
    static getFriendName() {
        return 'eggs';
    }
    /**
     * return the version of the package
     * @returns version example 8.0.0
     */
    static getPackageVersion() {
        return pjson.version;
    }
    /**
     * Get author name
     */
    static getAuthorName() {
        return 'Piero Proietti piero.proietti@gmail.com';
    }
    /**
     * Return the Debian version
     * @remarks to move in Utils
     * @returns {number} Versione di Debian
     */
    static getDebianVersion() {
        const cmd = "cat /etc/debian_version | /usr/bin/cut -f1 -d'.'";
        const version = Number(shx.exec(cmd, { silent: true }).stdout);
        return version;
    }
    /**
     * Return true if live system
     * @returns {boolean} isLive
     */
    static isLive() {
        let retVal = false;
        let paths = [
            '/lib/live/mount',
            '/lib/live/mount/rootfs/filesystem.squashfs',
            '/live/aufs' // mx-linux
        ];
        for (let i = 0; i < paths.length; i++) {
            if (Utils.isMountpoint(paths[i])) {
                retVal = true;
            }
        }
        return retVal;
    }
    /**
     * Ritorna vero se path è un mountpoint
     * @param path
     */
    static isMountpoint(path = '') {
        const cmd = `mountpoint -q ${path}`;
        // return 0 if the directory is a mountpoint, non-zero if not.
        const result = shx.exec(cmd, { silent: true }).code;
        return result === 0;
    }
    /**
     * return true if eggs run as root
     * @returns isRoot
     */
    static isRoot(command = '') {
        if (process.getuid && process.getuid() === 0) {
            return true;
        }
        Utils.warning(chalk.white(`${Utils.getFriendName()} ${command}`) + ` need to run with root privileges. Please, prefix it with sudo`);
        return false;
    }
    /**
     * get the kernel version
     */
    static kernerlVersion() {
        return os.release();
    }
    /**
     * return the name of network device
     */
    static netDeviceName() {
        const interfaces = Object.keys(os.networkInterfaces());
        let netDeviceName = '';
        for (const k in interfaces) {
            if (interfaces[k] != 'lo') {
                netDeviceName = interfaces[k];
            }
        }
        return netDeviceName;
    }
    /**
     * todo
     */
    static netAddress() {
        const { networkInterfaces } = require('os');
        const nets = networkInterfaces();
        const results = Object.create(null); // or just '{}', an empty object
        let address = '';
        for (const name of Object.keys(nets)) {
            for (const net of nets[name]) {
                // skip over non-ipv4 and internal (i.e. 127.0.0.1) addresses
                if (net.family === 'IPv4' && !net.internal) {
                    if (!results[name]) {
                        results[name] = [];
                    }
                    results[name].push(net.address);
                    address = net.address;
                }
            }
        }
        return address;
    }
    /**
     * todo
     */
    static netMasK() {
        const { networkInterfaces } = require('os');
        const nets = networkInterfaces();
        const results = Object.create(null); // or just '{}', an empty object
        let netmask = '';
        for (const name of Object.keys(nets)) {
            for (const net of nets[name]) {
                // skip over non-ipv4 and internal (i.e. 127.0.0.1) addresses
                if (net.family === 'IPv4' && !net.internal) {
                    if (!results[name]) {
                        results[name] = [];
                    }
                    results[name].push(net.address);
                    netmask = net.netmask;
                }
            }
        }
        return netmask;
    }
    /**
     * @returns dns
     */
    static netDns() {
        return dns.getServers();
    }
    /**
     * @returns gateway
     */
    static netGateway() {
        let cmd = `traceroute -m1 -n 8.8.8.8| grep ms| awk '{print $2}'`;
        let gw = shx.exec(cmd, { silent: true }).stdout.trim();
        return gw;
    }
    /**
     * userAdd
     * @param target
     * @param username
     * @param password
     * @param fullName
     */
    static userAdd(target = '/TARGET', username = 'live', password = 'evolution', fullName = '') {
        const cmd = `sudo chroot ${target} adduser ${username}\
                    --home /home/${username} \
                    --shell /bin/bash \
                    --disabled-password \
                    --gecos ${fullName},\
                            '',\
                            '',\
                            ''`;
        console.log(`addUser: ${cmd}`);
        shx.exec(cmd);
        const cmdPass = `echo ${username}:${password} | chroot ${target} chpasswd `;
        console.log(`addUser cmdPass: ${cmdPass}`);
        shx.exec(cmdPass);
        const cmdSudo = `chroot ${target} addgroup ${username} sudo`;
        console.log(`addUser cmdSudo: ${cmdSudo}`);
        shx.exec(cmdSudo, { silent: true });
    }
    /**
     * Return an array of the users of the system
     * @remarks to move in Utils
     * @returns {string[]} array di utenti
     */
    static usersList() {
        const out = shx.exec('/usr/bin/lslogins --noheadings -u -o user | grep -vw root', { silent: true }).stdout;
        const users = out.split('\n');
        return users;
    }
    /**
     * Create folder
     * @param dir
     * @param varius {recursive: true/false}
     */
    static shxMkDir(dir = '', varius) {
        if (varius.recursive) {
            shx.mkdir('-p', dir);
        }
        else {
            shx.mkdir(dir);
        }
    }
    /**
     * write a file
     * @param file
     * @param text
     */
    static write(file, text) {
        text = text.trim() + '\n';
        text = text.trim() + '\n';
        file = file.trim();
        fs.writeFileSync(file, text);
    }
    /**
     *
     * @param file
     * @param cmd
     */
    static writeX(file, cmd) {
        let text = `#!/bin/sh\n\n`;
        text += `# Created at: ${Utils.formatDate(new Date())}\n`;
        text += `# By: penguins_eggs v. ${Utils.getPackageVersion()}\n`;
        text += `# ==> Perri\'s Brewery edition <== \n\n`;
        text += cmd;
        Utils.write(file, text);
        shx.chmod('+x', file);
    }
    /**
     *
     * @param file
     * @param cmd
     */
    static writeXs(file, cmds) {
        let cmd = '';
        for (const elem of cmds) {
            cmd += elem + '\n';
        }
        Utils.writeX(file, cmd);
    }
    /**
     *
     * @param msg
     */
    static async customConfirm(msg = 'Select yes to continue... ') {
        const varResult = await Utils.customConfirmCompanion(msg);
        const result = JSON.parse(varResult);
        if (result.confirm === 'Yes') {
            return true;
        }
        else {
            return false;
        }
    }
    /**
     *
     * @param msg
     */
    static async customConfirmCompanion(msg = 'Select yes to continue... ') {
        return new Promise(function (resolve) {
            const questions = [
                {
                    type: 'list',
                    name: 'confirm',
                    message: msg,
                    choices: ['No', 'Yes'],
                    default: 'No'
                }
            ];
            inquirer.prompt(questions).then(function (options) {
                resolve(JSON.stringify(options));
            });
        });
    }
    /**
     *
     * @param msg
     */
    static async customConfirmAbort(msg = 'Select yes to continue, No to repeat, Abort to exit... ') {
        return new Promise(function (resolve) {
            const questions = [
                {
                    type: 'list',
                    name: 'confirm',
                    message: msg,
                    choices: ['No', 'Yes', 'Abort'],
                    default: 'No'
                }
            ];
            inquirer.prompt(questions).then(function (options) {
                resolve(JSON.stringify(options));
            });
        });
    }
    /**
     * titles
     * Penguin's are gettings alive!
     */
    static titles(command = '') {
        clear();
        console.log(chalk.cyan(figlet.textSync('eggs')));
        console.log(chalk.bgGreen.white('   ' + pjson.name + '   ') + chalk.bgWhite.blue(" Perri's Brewery edition ") + chalk.bgRed.whiteBright('    ver. ' + pjson.version + '   '));
        console.log('command: ' + chalk.bgBlack.white(command) + '\n');
    }
    /**
     *
     * @param verbose
     */
    static setEcho(verbose = false) {
        let echo = { echo: false, ignore: true };
        if (verbose) {
            echo = { echo: true, ignore: false };
        }
        return echo;
    }
}
exports.default = Utils;
